{
    
    Svg - A Real-time Scalable Vector Graphics implementation for the
    Cobra Programming Language.
    
    ---------------------------------------
    
    Copyright (c) 2010 Ben Golightly

    Permission is hereby granted, free of charge, to any person
    obtaining a copy of this software and associated documentation
    files (the "Software"), to deal in the Software without
    restriction, including without limitation the rights to use,
    copy, modify, merge, publish, distribute, sublicense, and/or sell
    copies of the Software, and to permit persons to whom the
    Software is furnished to do so, subject to the following
    conditions:

    The above copyright notice and this permission notice shall be
    included in all copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
    EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
    OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
    NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
    HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
    WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
    FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
    OTHER DEALINGS IN THE SOFTWARE.
    
    --------------------------------------- 

}
Unit
    Uses
        cobra2d,
        poxml  // 20100903 or later
        
Const
    
    (*
        SVG_VERSION
        Engine version 
    *)
    SVG_VERSION = 20100903 ; export

    (*
        SVG_EXTENSION_VENDOR
        Do not change this - it identifies the non-standard SVG extensions
        added to SVG metadata to support the features of this engine. 
    *)
    SVG_EXTENSION_VENDOR    = "tophatstuff"
    
    (*
        SVG_EXTENSION_VARIANT
        If you make any changes to this engine which break compatability,
        please change this string. e.g. to "svg/your-name/version". The
        engine will only load the extended data from SVG metadata when
        the value in the SVG file matches the value below.
    *)
    SVG_EXTENSION_VARIANT   = "svg/ths/1.0"
    
    (*
        DEFAULT_POLYGON_POINTS
        How many point-pairs to prepare by default for each polygon memory
        structure. The structure dobules in size when there are more points
        than would fit into the structure. Setting this too low means that
        loading polygons takes longer. Setting this too high will waste RAM.
        
        The "real" size of this structure in bytes is:
            DEFAULT_POLYGON_POINTS * 2 * SIZEOF_REAL
        e.g. 128 point-pairs is 1024 bytes = 1 kb
        
    *)
    DEFAULT_POLYGON_POINTS = 512
    
    (*
        Platform Specific:
    *)
    
    SIZEOF_REAL = 4
    SIZEOF_INT  = 4
    INFINITY = 2147483647 // 2^31 - 1
     

Type svgStrings = Record
    s: String
EndType ; export

Type svgPolygons = Record

    // Properties
    id              : ^svgStrings
    display         : Boolean
    fill            : Integer   // Color
    fill2           : Integer   // Color
    fillAngle       : Real
    fillOpacity     : Real
    stroke          : Integer   // Color
    strokeWidth     : Real
    strokeOpacity   : Real
    xShowPoints     : Boolean

    // Points
    points          : Element   // Raw points
    tPoints         : Element   // Transformed Points
    oPoints         : Element   // Original Raw Points
    totalPoints     : Integer
    pointsSize      : Integer   // Size of points bank
    updated         : Boolean
    minx, miny, maxx, maxy: Real // Min/Max coordinate offsets
    maxxd, maxyd: Real           // Difference between Max/min     
    
EndType ; export

Type svgs = Record
    x, y                : Real      // Linear transformation
    angle               : Real      // Rotation
    xscale              : Real
    yscale              : Real
    handlex, handley    : Real      // Rotation is done about the handle
    polygonList         : List of svgPolygons
    updated             : Boolean
EndType ; export

Var
    svgList: List of svgs
    
    svgQuality: Integer = 4 ; export
    svgCanvasX, svgCanvasY: Real = 0.0 ; export
    svgCanvasScaleX, svgCanvasScaleY: Real = 1.0 ; export

Function LoadSvg(fn: String) : ^svgs ; export
Var
    auto_polyId: Static Integer = -1
    
    // Xml
    svgFile: ^poxFiles
    rootNode: ^poxNodes
    svgNode: ^poxNodes
    childNode: ^poxNodes
    extensionNode: ^poxNodes
    extensionChildNode: ^poxNodes
    
    // Svg
    mySvg: ^svgs
    activePolygon: ^svgPolygons
    
    i, j, c, c2, points: Integer
    index, index1, index2: Integer
    s: String
    
    extensionsEnabled: Boolean = FALSE
Begin
    
    // Default SVG settings
    mySvg = NewItem(svgList)
    mySvg.x = 0.0
    mySvg.y = 0.0
    mySvg.handlex = 0.0
    mySvg.handley = 0.0
    mySvg.angle = 0.0
    mySvg.xscale = 1.0
    mySvg.yscale = 1.0
    mySvg.updated = TRUE  
    
    // Transverse XML
    svgFile = Pox_Load(fn)
    rootNode = PoxRootNode(svgFile)
    svgNode = PoxChild(rootNode, PoxChildIndexByName(rootNode, 'svg'))
    c = PoxChildren(svgNode)
    
    For i = 0 to (c-1)
        childNode = poxChild(svgNode, i)
        s = Lower(childNode.name.s)

        Conditions
        
            (s = 'polygon'):
                (*
                    <polygon id = "identifier"
                        display         = "block" OR "inline"
                        fill            = "#000000" TO "#ffffff"
                        fill-opacity    = "0.0" TO "1.0"
                        stroke          = "#000000" TO "#ffffff"
                        stroke-width    = "0" TO "100"
                        stroke-opacity  = "0.0" TO "1.0"
                        points          = "x1,y1, x2,y2, x3,y3, x4,y4 [...]"
                    />
                *) 
                Inc(auto_polyId)
                
                // New Polygon Defaults
                activePolygon = NewItem(mySvg.polygonList)
                New(activePolygon.id)
                activePolygon.id.s          = "_polygon"+auto_polyId+""
                activePolygon.display       = TRUE
                activePolygon.fill          = 0
                activePolygon.fill2         = 0
                activePolygon.fillAngle     = 0.0
                activePolygon.fillOpacity   = 1.0
                activePolygon.stroke        = 0
                activePolygon.strokeWidth   = 1.0
                activePolygon.strokeOpacity = 1.0
                
                activePolygon.pointsSize    = DEFAULT_POLYGON_POINTS * SIZEOF_REAL * 2
                activePolygon.points        = CreateBank(activePolygon.pointsSize)
                activePolygon.tPoints       = CreateBank(activePolygon.pointsSize)
                activePolygon.oPoints       = CreateBank(activePolygon.pointsSize)
                activePolygon.totalPoints   = 0
                activePolygon.updated       = TRUE
                
                // Polygon.id
                index = PoxParameterIndexByName(childNode, 'id')
                If index > -1 then activePolygon.id.s = PoxParameter(childNode, index)
                
                // Polygon.display
                index = PoxParameterIndexByName(childNode, 'display')
                If index > -1 then
                    s = Lower(PoxParameter(childNode, index))
                    activePolygon.display = (s = 'block')
                Endif
                
                // Polygon.fill
                index = PoxParameterIndexByName(childNode, 'fill')
                If index > -1 then
                    s = PoxParameter(childNode, index)
                    If (Svg_StringIsHex(s)) then
                        activePolygon.fill = HexStringToInt(s)
                        activePolygon.fill2 = activePolygon.fill
                    Endif
                Endif
                
                // Polygon.fillOpacity
                index = PoxParameterIndexByName(childNode, 'fill-opacity')
                If index > -1 then
                    activePolygon.fillOpacity = SafeToReal(PoxParameter(childNode, index))
                Endif
                
                // Polygon.stroke
                index = PoxParameterIndexByName(childNode, 'stroke')
                If index > -1 then
                    s = PoxParameter(childNode, index)
                    If (Svg_StringIsHex(s)) then activePolygon.stroke = HexStringToInt(s)
                Endif

                // Polygon.strokeWidth
                index = PoxParameterIndexByName(childNode, 'stroke-width')
                If index > -1 then
                    s = PoxParameter(childNode, index)
                    If (Svg_StringIsFloat(s)) then activePolygon.strokeWidth = SafeToReal(s)
                Endif
                
                // Polygon.strokeOpacity
                index = PoxParameterIndexByName(childNode, 'stroke-opacity')
                If index > -1 then
                    s = PoxParameter(childNode, index)
                    If (Svg_StringIsFloat(s)) then activePolygon.strokeOpacity = SafeToReal(s)
                Endif

                // Polygon.points
                index = PoxParameterIndexByName(childNode, 'points')
                If index > -1 then
                    s = PoxParameter(childNode, index)
                    ParseSvgPoints(s, activePolygon, activePolygon.pointsSize, activePolygon.totalPoints)                    
                Endif
                CopyBank(activePolygon.points, activePolygon.tPoints)
                CopyBank(activePolygon.points, activePolygon.oPoints)
                
                
            (s = 'metadata'):
                (*
                    <metadata>
                        <extension vendor="tophatstuff.co.uk" variant="ths/svg/1.0">
                            <origin x="0" y="0" />
                            <extends target="id" [...] />
                        </extension>
                    </metadata>
                *)
                            
                index = PoxChildIndexByName(childNode, 'extension')
                
                // If extension node exists
                If index > -1 then
                                
                    extensionsEnabled = FALSE
                    extensionNode = PoxChild(childNode, index)
                    index1 = PoxParameterIndexByName(extensionNode, 'vendor')
                    index2 = PoxParameterIndexByName(extensionNode, 'variant')
                    
                    // If vendor parameter exists
                    If (index1 > -1) then 
                        s = PoxParameter(extensionNode, index1)
                        If (s = SVG_EXTENSION_VENDOR) then
                        
                            // If variant parameter exists
                            If (index2 > -1) then 
                                s = PoxParameter(extensionNode, index2)
                                If (s = SVG_EXTENSION_VARIANT) then extensionsEnabled = TRUE
                            Endif
                            
                        Endif
                    Endif
                
                    // If vendor and variant are valid, we should scan through these child nodes!
                    If extensionsEnabled then
                    
                        c2 = PoxChildren(extensionNode)
                        For j = 0 to (c2-1)
                            extensionChildNode = poxChild(extensionNode, j)
                            s = Lower(extensionChildNode.name.s)
                            
                            Conditions
                                
                                (s = 'origin'):
                                    // <origin x="0" y="0" />
                                    
                                    // x / svg.handlex
                                    index = PoxParameterIndexByName(extensionChildNode, 'x')
                                    If (index > -1) then
                                        s = PoxParameter(extensionChildNode, index)
                                        If (Svg_StringIsFloat(s)) then mySvg.handlex = SafeToReal(s)
                                    Endif
                                    
                                    // y / svg.handley
                                    index = PoxParameterIndexByName(extensionChildNode, 'y')
                                    If (index > -1) then
                                        s = PoxParameter(extensionChildNode, index)
                                        If (Svg_StringIsFloat(s)) then mySvg.handley = SafeToReal(s)
                                    Endif
                                    
                                
                                (s = 'extends'):
                                
                                    // <extends target="star" fill2="#ffcc00" fill-angle="0" x-show-points="true" />
                                    
                                    // target / activePolygon
                                    index = PoxParameterIndexByName(extensionChildNode, 'target')
                                    If (index > -1) then
                                        s = PoxParameter(extensionChildNode, index)
                                        If PolygonExistsWithId(mySvg, s) then activePolygon = GetPolygonById(mySvg, s)
                                    Endif
                                    
                                    // polygon.fill2
                                    index = PoxParameterIndexByName(extensionChildNode, 'fill2')
                                    If (index > -1) then
                                        s = PoxParameter(extensionChildNode, index)
                                        If (Svg_StringIsHex(s)) then activePolygon.fill2 = HexStringToInt(s)
                                    Endif
                                    
                                    // polygon.fillAngle
                                    index = PoxParameterIndexByName(extensionChildNode, 'fill-angle')
                                    If (index > -1) then
                                        s = PoxParameter(extensionChildNode, index)
                                        If (Svg_StringIsFloat(s)) then activePolygon.fillAngle = SafeToReal(s)
                                    Endif
                                    
                                    // polygon.xShowPoints
                                    index = PoxParameterIndexByName(extensionChildNode, 'x-show-points')
                                    If (index > -1) then
                                        s = Lower(PoxParameter(extensionChildNode, index))
                                        activePolygon.xShowPoints = (s = 'true')
                                    Endif
                                
                                
                                Default: SvgError("Unkown extension command '"+s+"'.")
                                
                            EndConditions
                        
                        Next
                    
                    Endif
                    
                Endif
                
            default:
        EndConditions
                        
    Next
    
    PoxFree(svgFile)
    result = mySvg   
End

Procedure RenderSvgs(canvas: Element) ; export
Var
    mySvg: ^svgs
Begin

    Loop mySvg through svgList 
        RenderSvg(mySvg, canvas)
    EndLoop

End

Procedure RenderSvg(mySvg: ^svgs, canvas: Element) ; export
Var
    polygon: ^svgPolygons
Begin

    Loop polygon through mySvg.polygonList 
        If polygon.display then RenderSvgPolygon(mySvg, polygon, canvas)
    EndLoop
    
    mySvg.updated = FALSE
End

Var
    nodeX: Array [512] of Integer // Temporary array for storing scan-Line intersection x-coordinates

Procedure RenderSvgPolygon(mySvg: ^svgs, polygon: ^svgPolygons, canvas: Element)
Var
    i, k, yl: Integer // That's yL, not y1
    
    // Points
    px, py, px1, py1, px2, py2: Real
    ptx, pty: Real
    dx, dy, x, y, h: Real
    theta, thetaStart, theta2: Real
    
    // Fill
    fillAlpha: Real
    dy1, dy2: Real
    x1, x2, tmpx: Real
    pendown: Boolean
    nodes: Integer
    offset: Integer
    color: Integer
    r1, g1, b1, r2, g2, b2, r3, g3, b3, a1, a2, d: Real
    xthickness, ythickness: Real
    
Begin

    // Update polygon transformation
    // --------------------------------------------------------------
    If (polygon.updated) or (mySvg.updated) then
        polygon.updated = FALSE
        
        // Transform points
        For i = 0 to (polygon.totalPoints-1)
            
            // Get a copy of the start location
            px = PolyX(polygon.points, i)
            py = PolyY(polygon.points, i)
            
            // Difference from origin
            dx = px - mySvg.handlex
            dy = py - mySvg.handley
            
            // Rotate
            If mySvg.angle = 0 then
                x = dx
                y = dy
            Else
            
                h = Sqrt((dx*dx) +(dy*dy))
                
                // Start's current angle from origin
                thetaStart = Atan2(dx, dy)
                
                // New angle is mySvg.angle + thetaStart
                theta = -mySvg.angle + thetaStart
                
                // Given this angle, one right angle and the hypotenuse...
                // find new values for ptx and pty
                theta2 = 90 - theta
                
                x = h * Sin(theta)
                y = h * Sin(theta2)
                
            Endif
            
            // Take svg.scale and svg.transformation into account
            x = ((x * mySvg.xscale) + mySvg.x) 
            y = ((y * mySvg.yscale) + mySvg.y)
            
            // Apply canvas scaling
             x = (x * svgCanvasScaleX) + svgCanvasX
             y = (y * svgCanvasScaleY) + svgCanvasY
            
            // Apply!
            SetPolyX(polygon.tpoints, i, x)
            SetPolyY(polygon.tpoints, i, y)

        Next
        
        // Max/Min coords
        polygon.minx = PolyMinX(polygon.tpoints, polygon.totalPoints)
        polygon.miny = PolyMinY(polygon.tpoints, polygon.totalPoints)
        polygon.maxx = PolyMaxX(polygon.tpoints, polygon.totalPoints)
        polygon.maxy = PolyMaxY(polygon.tpoints, polygon.totalPoints)
        polygon.maxxd = polygon.maxx - polygon.minx
        polygon.maxyd = polygon.maxy - polygon.miny 
    
    Endif
        
    // Render the polygon fill
    // --------------------------------------------------------------
    If polygon.fillOpacity > 0 then
        fillAlpha = polygon.fillOpacity * 255.0
            
        (*
            To fill a polygon, cast horizontal scan-lines.
            For each Line made by two points, count intersections.
            Where the number of intersections are odd, the scan-line
            is inside the polygon and should fill it.

            In this implementation:
                
                For each scan line
                    Loop through all lines made by two points
                        If the Line intersects the scan line
                            Store the x-coordinate
                        Endif
                    EndLoop
                        
                    For each x pixel
                        If xpixel is a stored intersection, penDown = not penDown
                    Next
                Next                  
                
        *)
            
        // For each scan line
        For yl = polygon.miny to polygon.maxy step svgQuality
            nodes = 0
            penDown = FALSE

            // Loop through points, checking to see if created lines
            // intersect the scan line
            // !! Calls to PolyX/PolyY inlined for speed.
            For i = 0 to (polygon.totalPoints-1)
            
                // First set of points
                offset = i * SIZEOF_REAL * 2
                
                px1 = PeekReal(polygon.tPoints, offset)
                py1 = PeekReal(polygon.tPoints, offset + SIZEOF_REAL)

                // Second set of points
                If i < (polygon.totalPoints-1) then
                    offset = (i+1) * SIZEOF_REAL * 2
                Else
                    offset = 0
                Endif
                
                px2 = PeekReal(polygon.tPoints, offset)
                py2 = PeekReal(polygon.tPoints, offset + SIZEOF_REAL)
                        
                // Difference between point and scan line
                dy1 = py1 - yl
                dy2 = py2 - yl
                            
                // There's no intersection if the points are *both* below or *both* above the scan-line
                // Hence, if this is false there must be one above and one below,
                // Hence, an intersection
                If not ( ((dy1<0.0) and (dy2<0.0)) or ((dy1>=0.0) and (dy2>=0.0)) ) then
                
                    // Calculate & store position of intersection...
                    nodeX[nodes] = ToInt(px2 + (px1-px2)*dy2/(dy2-dy1)) // magic
                    Inc(nodes)    
                    
                Endif
                     
            Next
                
            // Draw!
            For i = (polygon.minx-1) to (polygon.maxx+1)
                For k = 0 to nodes-1
                    If i = nodeX[k] then penDown = not penDown
                Next
                
                If penDown and ((i mod svgQuality) = 0) then
                    
                    // Fill color
                    If (polygon.fill = polygon.fill2) or (polygon.maxyd < 1) then
                        
                        // Standard 
                        color = IntToRGB(polygon.fill, polygon.fillOpacity)
                        
                    Else
                    
                        // Gradient
                        // ToDo: Take fillAngle into consideration!
                        
                        d = ToReal(yl-polygon.miny) 
                    
                        // Color 1
                        r1 = polygon.fill AND $FF0000
                        g1 = polygon.fill AND $00FF00
                        b1 = polygon.fill AND $0000FF
                        r1 = r1 / (256*256)
                        g1 = g1 / (256)
                        a1 = 1 - (d / polygon.maxyd)
                    
                        // Color 2
                        r2 = polygon.fill2 AND $FF0000
                        g2 = polygon.fill2 AND $00FF00
                        b2 = polygon.fill2 AND $0000FF
                        r2 = r2 / (256*256)
                        g2 = g2 / (256)
                        a2 = d / polygon.maxyd
                        
                        // Mix
                        r3 = (r1 * a1) + (r2 * a2)
                        g3 = (g1 * a1) + (g2 * a2)
                        b3 = (b1 * a1) + (b2 * a2)
                        
                        color = ToRGBA(r3, g3, b3, 255*polygon.fillOpacity)
                    
                    Endif
                    
                    // Draw!
                    Rect(i-(svgQuality/2), yl-(svgQuality/2), svgQuality, svgQuality, color, TRUE, canvas)
                Endif
            Next
                
        Next                
    Endif

    // Render the polygon outline
    // --------------------------------------------------------------
    For i = 0 to (polygon.totalPoints-1)
        
        // Point 1
        px1 = PolyX(polygon.tPoints, i)
        py1 = PolyY(polygon.tPoints, i)
        
        // Point 2
        If i < (polygon.totalPoints-1) then
            px2 = PolyX(polygon.tPoints, i+1)
            py2 = PolyY(polygon.tPoints, i+1)
        Else
            px2 = PolyX(polygon.tPoints, 0)
            py2 = PolyY(polygon.tPoints, 0) 
        Endif
        
        // Line between two points
        xthickness = polygon.strokeWidth * mySvg.xscale * svgCanvasScaleX
        ythickness = polygon.strokeWidth * mySvg.yscale * svgCanvasScaleY
        Conditions
            (polygon.strokeWidth = 0): // No line
            
            ((xthickness <= 1) and (ythickness <=1)):
                // Thin line
                Line(px1, py1, px2, py2, IntToRGB(polygon.stroke, polygon.strokeOpacity), canvas)
                
            Default:
                // Thick line
                BLine(px1, py1, px2, py2, IntToRGB(polygon.stroke, polygon.strokeOpacity), canvas, xthickness, ythickness)
                
        EndConditions
        
        // Optionally draw a box at each point
        If polygon.xShowPoints then
            Rect(px1, py1, 1, 1, ToRGBA(255,255,255), FALSE, canvas)
            Rect(px1-1, py1-1, 3, 3, ToRGBA(0,0,0), FALSE, canvas)
        Endif
        
        
    Next   
    

End

Procedure BLine(x0: Integer, y0: Integer, x1: Integer, y1: Integer, color: Integer, canvas: Element, xthickness: Integer, ythickness: Integer)
Var
    steep: Boolean
    iTmp: Integer
    deltax, deltay: Integer
    error: Integer
    ystep: Integer
    x, y: Integer
    xthickOver2, ythickOver2: Integer
Begin
    // http://en.wikipedia.org/wiki/Bresenham%27s_line_algorithm
    // Instead of plotting pixels, this algo is modified to plot
    // rects of size (xthickness, ythickness)

    steep = abs(y1 - y0) > abs(x1 - x0) 
 
    If steep then
        
        // swap(x0, y0)
        iTmp = x0
        x0 = y0
        y0 = iTmp
        
        // swap(x1, y1)
        iTmp = x1
        x1 = y1
        y1 = iTmp        
        
    Endif
        
     If x0 > x1 then

         // swap(x0, x1)
        iTmp = x0
        x0 = x1
        x1 = iTmp
         
         // swap(y0, y1)
        iTmp = y0
        y0 = y1
        y1 = iTmp
         
     Endif
     
     deltax = x1 - x0
     deltay = abs(y1 - y0)
     error = deltax / 2
     ystep = 0
     y = y0
     
     If y0 < y1 then ystep = 1 else ystep = -1
     
     For x = x0 to x1
     
        xthickOver2 = xthickness/2
        ythickOver2 = ythickness/2
     
        If steep then
            Rect(y-ythickOver2, x-xthickOver2, xthickness, ythickness, color, TRUE, canvas)
        Else
            Rect(x-xthickOver2, y-ythickOver2, xthickness, ythickness, color, TRUE, canvas)
        Endif
        
        error = error - deltay
        If error < 0 then
            y = y + ystep
            error = error + deltax
        Endif
    Next

End

Procedure UpdateAllSvgs() ; export
Var
    mySvg: ^svgs
Begin
    Loop mySvg through svgList
        mySvg.updated = TRUE
    EndLoop
End

Procedure SetSvgCanvasScale(x: Real, y: Real) ; export
Begin
    If (svgCanvasScaleX = x) and (svgCanvasScaley = y) then exit
    svgCanvasScaleX = x
    svgCanvasScaleY = y
    UpdateAllSvgs()
End

Procedure SetSvgCanvasOffset(x: Real, y: Real) ; export
Begin
    If (svgCanvasX = x) and (svgCanvasY = y) then exit
    svgCanvasX = x
    svgCanvasY = y
    UpdateAllSvgs()
End

Function PolygonExistsWithId(mySvg: ^svgs, s: String) : Boolean ; export
Var
    polygon: ^svgPolygons
Begin
    result = FALSE

    Loop polygon through mySvg.polygonList
        If polygon.id.s = s then
            result = TRUE
            exit
        Endif
    EndLoop
End

Function GetPolygonById(mySvg: ^svgs, s: String) : ^svgPolygons ; export
Var
    polygon: ^svgPolygons
Begin
    Loop polygon through mySvg.polygonList
        If polygon.id.s = s then
            result = polygon
            exit
        Endif
    EndLoop
End

Procedure SetSvgHandle(mySvg: ^svgs, x: Real, y: Real) ; export
Begin
    If (mySvg.handlex = x) and (mySvg.handley = y) then exit
    mySvg.handlex = x
    mySvg.handley = y
    mySvg.updated = TRUE
End

Procedure SetSvgScale(mySvg: ^svgs, x: Real, y: Real) ; export
Begin
    mySvg.xscale = x
    mySvg.yscale = y
    mySvg.updated = TRUE
End

Procedure SetSvgAngle(mySvg: ^svgs, r: Real) ; export
Begin
    If (mySvg.angle = r) then exit
    mySvg.angle = r
    mySvg.updated = TRUE
End

Procedure SetSvgPosition(mySvg: ^svgs, x: Real, y: Real) ; export
Begin
    mySvg.x = x
    mySvg.y = y
    mySvg.updated = TRUE
End

Function PolyX(bank: Element, i: Integer) : Real ; export
Begin
    result = PeekReal(bank, i * (SIZEOF_REAL * 2))
End

Function PolyY(bank: Element, i: Integer) : Real ; export
Begin
    result = PeekReal(bank, (i * (SIZEOF_REAL * 2)) + SIZEOF_REAL)
End

Procedure SetPolyX(bank: Element, i: Integer, v: Real) ; export
Begin
    PokeReal(bank, i * (SIZEOF_REAL * 2), v)
End

Procedure SetPolyY(bank: Element, i: Integer, v: Real) ; export
Begin
    PokeReal(bank, (i * (SIZEOF_REAL * 2)) + SIZEOF_REAL, v)
End

Function PolyMinX(bank: Element, c: Integer) : Real
Var
    i: Integer
Begin
    result = INFINITY
    
    For i = 0 to (c-1)
        result = Min(result, PolyX(bank, i))
    Next
     
End

Function PolyMinY(bank: Element, c: Integer) : Real
Var
    i: Integer
Begin
    result = INFINITY
    
    For i = 0 to (c-1)
        result = Min(result, PolyY(bank, i))
    Next
     
End

Function PolyMaxX(bank: Element, c: Integer) : Real
Var
    i: Integer
Begin
    result = -INFINITY
    
    For i = 0 to (c-1)
        result = Max(result, PolyX(bank, i))
    Next
     
End

Function PolyMaxY(bank: Element, c: Integer) : Real
Var
    i: Integer
Begin
    result = -INFINITY
    
    For i = 0 to (c-1)
        result = Max(result, PolyY(bank, i))
    Next
     
End

// For Parsing SVG Coordinate point-pairs
Var
    parseSvgPoints_index0: Integer
    parseSvgPoints_index1: Integer
    parseSvgPoints_string: String
    parseSvgPoints_strlen: Integer
    parseSvgPoints_lexeme: String

Function ParseSvgPoints(s: String, polygon: ^svgPolygons, ByRef pointsSize: Integer, ByRef totalPoints: Integer) : Boolean
Var
    i: Integer
    c: String
    token: Integer
    active: Integer
    offset: Integer
    point1, point2: Real
Begin

    InitSvgPointsToken(s)
    active = -1

    Repeat
        token = GetNextSvgPointsToken()
        
        Case token of
             1: // Comma
                Case active of
                    -1:
                     1: active=2
                     2:
                EndCase
            
             2: // Float
                Case active of
                    -1: active = 1;  point1 = ToReal(parseSvgPoints_lexeme)
                     1:
                     2: active = -1; point2 = ToReal(parseSvgPoints_lexeme);
                     
                        // Got a full point-pair; let's process it
                        if ((totalPoints+1) * SIZEOF_REAL * 2) > pointsSize then
                            pointsSize = pointsSize * 2
                            FreeBank(polygon.points)
                            polygon.points = CreateBank(pointsSize)
                            totalPoints = 0
                            InitSvgPointsToken(s)
                        Else
                            offset = totalPoints * (SIZEOF_REAL * 2)
                            PokeReal(polygon.points, offset, point1)
                            PokeReal(polygon.points, offset + SIZEOF_REAL, point2)
                            Inc(totalPoints) 
                        Endif
                                               
                EndCase
            
            -1:
        EndCase
    
    Until token = -1
        
End

Function InitSvgPointsToken(s: String) : Boolean
Begin
    parseSvgPoints_index0 = 0
    parseSvgPoints_index1 = 0
    parseSvgPoints_string = s
    parseSvgPoints_strlen = Length(s)
    parseSvgPoints_lexeme = ""
End

Function GetNextSvgPointsToken() : Integer
(*
    Result:
        -1: End of Stream / error
         1: Comma
         2: Float value
*)
Var
    l: Integer
    c: String
Begin

    parseSvgPoints_index0 = parseSvgPoints_index1
    result = -1
    
    // Skip through whitespace
    While (TRUE)                
        c = Mid(parseSvgPoints_string, parseSvgPoints_index0+1, 1)
        If Svg_CharIsWhiteSpace(c) = FALSE then break
        Inc(parseSvgPoints_index0)
        
        If parseSvgPoints_index0 >= parseSvgPoints_strlen then break
    Wend
    
    parseSvgPoints_index1 = parseSvgPoints_index0
    
    // Scan through until a delimiter is met, incrementing index1    
    While (TRUE)    
        c = Mid(parseSvgPoints_string, parseSvgPoints_index1+1, 1)
        If Svg_CharIsPointsDelim(c) then break
        Inc(parseSvgPoints_index1)
        
        If parseSvgPoints_index1 >= parseSvgPoints_strlen then break
    Wend
    
    If (parseSvgPoints_index1 - parseSvgPoints_index0) = 0 then Inc(parseSvgPoints_index1)
    
    // Lexeme Isolatted, lies between index0 and index1       
    parseSvgPoints_lexeme = Mid(parseSvgPoints_string, parseSvgPoints_index0+1, (parseSvgPoints_index1-parseSvgPoints_index0))
    
    result = -1
    
    Conditions
        (parseSvgPoints_lexeme = ','): result = 1
        (Svg_StringIsFloat(parseSvgPoints_lexeme)): result = 2
    EndConditions    
    
End

Function Svg_CharIsPointsDelim(c: String) : Boolean
Begin
    result = FALSE
    If (c = ',') then result = TRUE  ; exit
    result = Svg_CharIsWhitespace(c)  
End

Function Svg_CharIsNumeric(c: String) : Boolean
Var
    a: Integer
Begin
    a = Asc(c)    
    If (a >= 48) and (a <= 57) then result = TRUE else result = FALSE    
End

Function Svg_CharIsWhitespace(c: String) : Boolean
Var
    a: Integer
Begin
    a = Asc(c) // Tab = 9, Space = 32, Newline = 10    
    If (a = 9) or (a = 32) or (a = 10) then result = TRUE else result = FALSE
End

Function Svg_CharIsHex(c: String) : Boolean
Var
    a: Integer
Begin
    a = Asc(c)
    result = FALSE    
    If (a >= 48) and (a <= 57) then result = TRUE ; exit
    If (a >= 65) and (a <= 70) then result = TRUE ; exit
    If (a >= 97) and (a <= 102) then result = TRUE ; exit    
End

Function Svg_StringIsInt(s: String) : Boolean ; export
Var
    l, i: Integer
    c: String
Begin
    result = FALSE
    l = Length(s)
    If l < 1 then exit
    
    For i = 1 to l
        c = Mid(s, i, 1)
        
        If Not (Svg_CharIsNumeric(c) or (c = '-')) then exit
        If (i <> 1) and (c = '-') then exit 
    Next
    
    result = TRUE        
End

Function Svg_StringIsFloat(s: String) : Boolean
Var
    l, i: Integer
    radix: Integer = 0
    c: String
Begin
    result = FALSE
    l = Length(s)
    If l < 1 then exit
    
    For i = 1 to l
        c = Mid(s, i, 1)
        
        If Not (Svg_CharIsNumeric(c) or (c = '-') or (c = '.')) then exit
        If (i <> 1) and (c = '-') then exit
        
        If (c = '.') then
            Inc(radix)
            If i = l then exit
        Endif
        If radix > 1 then exit 
    Next
    
    result = TRUE        
End

Function Svg_StringIsHex(s: String) : Boolean
Var
    l, i: Integer
    c: String
    sCount: Integer = 0 // symbolCount
Begin
    result = FALSE
    l = Length(s)
    If l < 2 then exit
    
    For i = 1 to l
        c = Mid(s, i, 1)
        
        If Not (Svg_CharIsHex(c) or (c = '-') or (c = '#')) then exit
        If (i > 1) and (c = '-') then exit
        If (i > 2) and (c = '#') then exit
        If (c = '#') then Inc(sCount)
    Next
    
    result = TRUE
    
    If sCount <> 1 then result = FALSE        
End

Function HexStringToInt(s: String): Integer
Var
    l, i, c: Integer
    polarity: Integer
    start: Integer
    total: Integer = 0
    s2: String
Begin
    l = Length(s)
    s = Upper(s)
    
    // Remove -, # (note assumes valid hex string, e.g. -#FF)
    If Left(s, 1) = '-' then
        polarity = -1
        start = 3
    Else
        polarity = 1
        start = 2
    Endif
    
    // convert #ABC to #AABBCC
    If (l-start+1) = 3 then
        s2 = "#"
        If polarity = -1 then s2 = "-"+s2
        s2 = s2 + Mid(s, start, 1) + Mid(s, start, 1)
        s2 = s2 + Mid(s, start+1, 1) + Mid(s, start+1, 1)
        s2 = s2 + Mid(s, start+2, 1) + Mid(s, start+2, 1)
        result = HexStringToInt(s2)
        exit
    Endif
    
    For i = start to l
        c = Asc(Mid(s, i, 1))
        If (c >= 48) and (c <= 57) then c = c - 48 // 0-9
        If (c >= 65) and (c <= 70) then c = c + 10 - 65 // A-F
        
        If (l - i) > 0 then
            total = total + (c * (16^(l - i)))
        Else
            total = total + c
        Endif         
    
    Next
    
    result = total * polarity   

End

Function IntToRGB(i: Integer, a: Real) : Integer
Var
    r, g, b: Integer
Begin

    r = i AND $FF0000
    g = i AND $00FF00
    b = i AND $0000FF
    
    r = r / (256*256)
    g = g / (256)

    result = ToRGBA(r,g,b, a*255)

End

Function SafeToReal(s: String): Real
Begin
    result = 0.0
    If Svg_StringIsFloat(s) then result = ToReal(s)
End   

Procedure CopyBank(source: Element, ByRef dest: Element)
Var
    sizeSource, sizeDest: Integer
    i, b: Integer
Begin
    sizeSource = BankSize(source)
    sizeDest   = BankSize(dest)
    
    If sizeSource > sizeDest then
        FreeBank(dest)
        dest = CreateBank(sizeSource)
    Endif
    
    For i = 0 to (sizeSource-1)
        b = PeekByte(source, i)
        PokeByte(dest, i, b)
    Next
    
End

Procedure SvgError(msg: String)
Begin
    MessageBox(msg, "Svg")
End

Begin
End